/* $Id: team.cpp,v 1.55 2004/05/18 03:28:23 jbm Exp $ */
/*
   Copyright (C) 2003 by David White <davidnwhite@optusnet.com.au>
   Part of the Battle for Wesnoth Project http://wesnoth.whitevine.net

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License.
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.

   See the COPYING file for more details.
*/

#include "game_config.hpp"
#include "replay.hpp"
#include "team.hpp"
#include "util.hpp"

#include <algorithm>
#include <cstdlib>
#include <sstream>

namespace {
	std::vector<team>* teams = NULL;
}

teams_manager::teams_manager(std::vector<team>& teams_list)
{
	teams = &teams_list;
}

teams_manager::~teams_manager()
{
	teams = NULL;
}

team::target::target(const config& cfg)
              : criteria(cfg), value(atof(cfg["value"].c_str()))
{
}

void team::target::write(config& cfg) const
{
	cfg = criteria;
}

team::team_info::team_info(const config& cfg)
{
	gold = cfg["gold"];
	income = cfg["income"];
	name = cfg["name"];
	team_name = cfg["team_name"];
	if(team_name.empty())
		team_name = cfg["side"];

	description = cfg["description"];

	const std::string& village_income = cfg["village_gold"];
	if(village_income.empty())
		income_per_village = game_config::village_income;
	else
		income_per_village = atoi(village_income.c_str());

	const std::string& aggression_val = cfg["aggression"];
	if(aggression_val.empty())
		aggression = 0.5;
	else
		aggression = atof(aggression_val.c_str());

	const std::string& caution_val = cfg["caution"];
	if(caution_val.empty())
		caution = 0.25;
	else
		caution = atof(caution_val.c_str());

	const std::string& enemies_list = cfg["enemy"];
	if(!enemies_list.empty()) {
		std::vector<std::string> venemies = config::split(enemies_list);
		for(std::vector<std::string>::const_iterator i = venemies.begin();
		    i != venemies.end(); ++i) {
			enemies.push_back(atoi(i->c_str()));
		}
	}

	if(cfg["controller"] == "human")
		controller = HUMAN;
	else if(cfg["controller"] == "network")
		controller = NETWORK;
	else
		controller = AI;

	ai_algorithm = cfg["ai_algorithm"];


	const config::child_list& ai_parameters = cfg.get_children("ai");
	for(config::child_list::const_iterator aiparam = ai_parameters.begin(); aiparam != ai_parameters.end(); ++aiparam) {
		ai_params.append(**aiparam);
	}

	const std::string& scouts_val = cfg["villages_per_scout"];
	if(scouts_val.empty()) {
		villages_per_scout = 4;
	} else {
		villages_per_scout = atoi(scouts_val.c_str());
	}

	const std::string& leader_val = cfg["leader_value"];
	if(leader_val.empty()) {
		leader_value = 3.0;
	} else {
		leader_value = atof(leader_val.c_str());
	}

	const std::string& village_val = cfg["village_value"];
	if(village_val.empty()) {
		village_value = 1.0;
	} else {
		village_value = atof(village_val.c_str());
	}

	std::vector<std::string> recruits = config::split(cfg["recruit"]);
	for(std::vector<std::string>::const_iterator i = recruits.begin(); i != recruits.end(); ++i) {
		can_recruit.insert(*i);
	}

	recruitment_pattern = config::split(cfg["recruitment_pattern"]);

	//default recruitment pattern is to buy 2 fighters for every 1 archer
	if(recruitment_pattern.empty()) {
		recruitment_pattern.push_back("fighter");
		recruitment_pattern.push_back("fighter");
		recruitment_pattern.push_back("archer");
	}

	//additional targets
	for(config::const_child_itors tgts = cfg.child_range("target");
	    tgts.first != tgts.second; ++tgts.first) {
		targets.push_back(target(**tgts.first));
	}

	use_shroud = (cfg["shroud"] == "yes");
	use_fog = (cfg["fog"] == "yes");
	share_maps = (cfg["share_maps"] == "yes");
	share_vision = (cfg["share_vision"] == "yes");

	std::cerr << "use_shroud=" << use_shroud << ",use_fog=" << use_fog;
	std::cerr << ",share_maps=" << share_maps << ",share_vision=" << share_vision << '\n';
	
	music = cfg["music"];
}

void team::team_info::write(config& cfg) const
{
	cfg["gold"] = gold;
	cfg["income"] = income;
	cfg["name"] = name;
	cfg["team_name"] = team_name;
	cfg["description"] = description;

	char buf[50];
	sprintf(buf,"%d",income_per_village);
	cfg["village_gold"] = buf;

	sprintf(buf,"%f",aggression);
	cfg["aggression"] = buf;

	std::stringstream enemies_str;
	for(std::vector<int>::const_iterator en = enemies.begin(); en != enemies.end(); ++en) {
		enemies_str << *en;
		if(en+1 != enemies.end())
			enemies_str << ",";
	}

	cfg["enemy"] = enemies_str.str();

	switch(controller) {
	case AI: cfg["controller"] = "ai"; break;
	case HUMAN: cfg["controller"] = "human"; break;
	case NETWORK: cfg["controller"] = "network"; break;
	default: assert(false);
	}

	sprintf(buf,"%d",villages_per_scout);
	cfg["villages_per_scout"] = buf;

	sprintf(buf,"%f",leader_value);
	cfg["leader_value"] = buf;

	sprintf(buf,"%f",village_value);
	cfg["village_value"] = buf;

	for(std::vector<target>::const_iterator tg = targets.begin(); tg != targets.end(); ++tg) {
		tg->write(cfg.add_child("target"));
	}

	std::stringstream can_recruit_str;
	for(std::set<std::string>::const_iterator cr = can_recruit.begin(); cr != can_recruit.end(); ++cr) {
		if(cr != can_recruit.begin())
			can_recruit_str << ",";

		can_recruit_str << *cr;
	}

	cfg["recruit"] = can_recruit_str.str();

	std::stringstream recruit_pattern_str;
	for(std::vector<std::string>::const_iterator p = recruitment_pattern.begin(); p != recruitment_pattern.end(); ++p) {
		if(p != recruitment_pattern.begin())
			recruit_pattern_str << ",";

		recruit_pattern_str << *p;
	}

	cfg["recruitment_pattern"] = recruit_pattern_str.str();

	cfg["shroud"] = use_shroud ? "yes" : "no";
	cfg["fog"] = use_fog ? "yes" : "no";
	cfg["share_maps"] = share_maps ? "yes" : "no";
	cfg["share_vision"] = share_vision ? "yes" : "no";

	if(music.empty() == false)
		cfg["music"] = music;
}

team::team(const config& cfg, int gold) : gold_(gold), auto_shroud_updates_(true), info_(cfg)
{
	//gold is the maximum of 'gold' and what is given in the config file
	if(info_.gold.empty() == false)
		gold_ = maximum(gold,::atoi(info_.gold.c_str()));

	//load in the villages the side controls at the start
	const config::child_list& villages = cfg.get_children("village");
	for(config::child_list::const_iterator v = villages.begin(); v != villages.end(); ++v) {
		villages_.insert(gamemap::location(**v));
	}

	const std::string& shroud_data = cfg["shroud_data"];
	for(std::string::const_iterator sh = shroud_data.begin(); sh != shroud_data.end(); ++sh) {
		if(*sh == '|')
			shroud_.resize(shroud_.size()+1);

		if(shroud_.empty() == false) {
			if(*sh == '1')
				shroud_.back().push_back(true);
			else if(*sh == '0')
				shroud_.back().push_back(false);
		}
	}
}

void team::write(config& cfg) const
{
	info_.write(cfg);

	char buf[50];
	sprintf(buf,"%d",gold_);
	cfg["gold"] = buf;

	//write village locations
	for(std::set<gamemap::location>::const_iterator t = villages_.begin(); t != villages_.end(); ++t) {
		t->write(cfg.add_child("village"));
	}

	std::stringstream shroud_str;
	for(std::vector<std::vector<bool> >::const_iterator sh = shroud_.begin(); sh != shroud_.end(); ++sh) {
		shroud_str << '|';

		for(std::vector<bool>::const_iterator i = sh->begin(); i != sh->end(); ++i) {
			shroud_str << (*i ? '1' : '0');
		}

		shroud_str << '\n';
	}

	cfg["shroud_data"] = shroud_str.str();
}

void team::get_village(const gamemap::location& loc)
{
	villages_.insert(loc);
}

void team::lose_village(const gamemap::location& loc)
{
	if(owns_village(loc)) {
		villages_.erase(villages_.find(loc));
	}
}

void team::clear_villages()
{
	villages_.clear();
}

const std::set<gamemap::location>& team::villages() const
{
	return villages_;
}

bool team::owns_village(const gamemap::location& loc) const
{
	return villages_.count(loc) > 0;
}

int team::gold() const
{
	return gold_;
}

int team::income() const
{
	return atoi(info_.income.c_str()) +
	       villages_.size()*info_.income_per_village+game_config::base_income;
}

void team::new_turn()
{
	gold_ += income();
}

void team::spend_gold(int amount)
{
	gold_ -= amount;
}

void team::set_income(int amount)
{
	info_.income = lexical_cast<std::string>(amount);
}

const std::set<std::string>& team::recruits() const
{
	return info_.can_recruit;
}

std::set<std::string>& team::recruits()
{
	return info_.can_recruit;
}

const std::vector<std::string>& team::recruitment_pattern() const
{
	return info_.recruitment_pattern;
}

const std::string& team::name() const
{
	return info_.name;
}

bool team::is_enemy(int n) const
{
	const size_t index = size_t(n-1);
	if(teams == NULL || index >= teams->size()) {
		return false;
	}

	//we're not enemies of ourselves
	if(&(*teams)[index] == this) {
		return false;
	}

	//if we have a team name, we are friends with anyone who has the same team name
	if(info_.team_name.empty() == false) {
		return (*teams)[index].info_.team_name != info_.team_name;
	}

	//if enemies aren't listed, then everyone is an enemy
	if(info_.enemies.empty())
		return true;

	return std::find(info_.enemies.begin(),info_.enemies.end(),n) != info_.enemies.end();
}

double team::aggression() const
{
	return info_.aggression;
}

double team::caution() const
{
	return info_.caution;
}

bool team::is_human() const
{
	return info_.controller == team_info::HUMAN;
}

bool team::is_ai() const
{
	return info_.controller == team_info::AI;
}

bool team::is_network() const
{
	return info_.controller == team_info::NETWORK;
}

void team::make_human()
{
	info_.controller = team_info::HUMAN;
}

void team::make_ai()
{
	info_.controller = team_info::AI;
}

const std::string& team::team_name() const
{
	return info_.team_name;
}

void team::change_team(const std::string& name)
{
	info_.team_name = name;
}

const std::string& team::ai_algorithm() const
{
	return info_.ai_algorithm;
}

const config& team::ai_parameters() const
{
	return info_.ai_params;
}

void team::make_network()
{
	info_.controller = team_info::NETWORK;
}

double team::leader_value() const
{
	return info_.leader_value;
}

double team::village_value() const
{
	return info_.village_value;
}

int team::villages_per_scout() const
{
	return info_.villages_per_scout;
}

std::vector<team::target>& team::targets()
{
	return info_.targets;
}

bool team::shrouded(size_t x, size_t y) const
{
	bool res = side_shrouded(x,y);
	
	if(!res || teams == NULL || info_.team_name == "" || !info_.share_maps)
		return res;

	for(std::vector<team>::const_iterator t = teams->begin(); t != teams->end(); ++t) {
		if(t->info_.team_name != info_.team_name)
			continue;
		if(t->uses_shroud() && !t->side_shrouded(x,y))
			return false;
	}
	return true;
}

bool team::fogged(size_t x, size_t y) const
{
	bool res = side_fogged(x,y);
	
	if(!res || !teams || info_.team_name.empty() || !info_.share_maps || !info_.share_vision) {
		return res;
	}

	for(std::vector<team>::const_iterator t = teams->begin(); t != teams->end(); ++t) {
		if(t->info_.team_name != info_.team_name)
			continue;
		if(t->uses_fog() && !t->side_fogged(x,y))
			return false;
	}
	return true;
}

bool team::side_shrouded(size_t x, size_t y) const
{
	if(info_.use_shroud == false)
		return false;

	if(x >= shroud_.size())
		return true;

	if(y >= shroud_[x].size())
		return true;

	return !shroud_[x][y];
}

bool team::clear_shroud(size_t x, size_t y)
{
	if(info_.use_shroud == false)
		return false;

	if(x >= shroud_.size())
		shroud_.resize(x+1);

	if(y >= shroud_[x].size())
		shroud_[x].resize(y+1);

	if(shroud_[x][y] == false) {
		shroud_[x][y] = true;
		return true;
	} else {
		return false;
	}
}

bool team::side_fogged(size_t x, size_t y) const
{
	if(info_.use_fog == false)
		return side_shrouded(x,y);

	if(x >= fog_.size())
		return true;

	if(y >= fog_[x].size())
		return true;

	if(fog_[x][y])
		return side_shrouded(x,y);
	else
		return true;
}

bool team::clear_fog(size_t x, size_t y)
{
	if(info_.use_fog == false)
		return false;

	if(x >= fog_.size())
		fog_.resize(x+1);

	if(y >= fog_[x].size())
		fog_[x].resize(y+1);

	if(fog_[x][y] == false) {
		fog_[x][y] = true;
		return true;
	} else {
		return false;
	}
}

void team::refog()
{
	for(std::vector<std::vector<bool> >::iterator i = fog_.begin();
	    i != fog_.end(); ++i) {
		std::fill(i->begin(),i->end(),false);
	}
}


bool team::knows_about_team(size_t index) const
{
	const team& t = (*teams)[index];
	//We know about our team
	if(this == &t) return true;
	/*
	We ask two questions to determine if we know this team:
	1) can we see this team through the shroud?
	2) can we see this team through the fog?
	If both answers are "yes" then we know about this team.
	If we aren't using fog or shroud, the answer to that question is automatically
	*/
	bool shroud = !uses_shroud() || !is_enemy(index+1) && t.uses_shroud() && t.share_maps();
	bool fog = !uses_fog() || !is_enemy(index+1) && t.uses_fog() && t.share_vision();
	return shroud && fog;
}

const std::string& team::music() const
{
	return info_.music;
}

int team::nteams()
{
	if(teams == NULL) {
		return 0;
	} else {
		return teams->size();
	}
}

const std::set<gamemap::location> vacant_villages(const std::set<gamemap::location>& villages, const unit_map& units)
{
	std::set<gamemap::location> res;
	
	for(std::set<gamemap::location>::const_iterator i = villages.begin(); i != villages.end(); ++i) {
		if(units.count(*i) == 0) {
			res.insert(*i);
		}
	}

	return res;
}

bool is_observer()
{
	if(teams == NULL) {
		return true;
	}

	for(std::vector<team>::const_iterator i = teams->begin(); i != teams->end(); ++i) {
		if(i->is_human()) {
			return false;
		}
	}

	return true;
}

void validate_side(int side)
{
	if(teams == NULL) {
		return;
	}

	if(side < 1 || side > int(teams->size())) {
		std::cerr << "invalid side " << side << " throwing game_error\n";
		throw gamestatus::game_error("invalid side found in unit definition");
	}
}
